@@ -26,6 +26,12 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
+def cast_tuple(val, size = 1):
+    return val if isinstance(val, tuple) else (val,) * size
+
+def calc_same_padding(kernel_size, dilation = 1):
+    return dilation * (kernel_size - 1) // 2
+
 # keyword argument helpers
 
 def pick_and_pop(keys, d):
@@ -549,7 +555,6 @@ class Sparse3DNA(nn.Module):
         query_num_frames_chunk = None
     ):
         super().__init__()
-        assert kernel_size % 2 == 1, 'kernel size must be odd'
         inner_dim = dim_head * heads
         self.heads = heads
         self.scale = dim_head ** -0.5
@@ -559,11 +564,24 @@ class Sparse3DNA(nn.Module):
         self.to_kv = nn.Linear(dim, inner_dim * 2, bias = False)
         self.to_out = nn.Linear(inner_dim, dim)
 
-        self.dilation = dilation
-        self.kernel_size = kernel_size
-        self.video_padding = dilation * (kernel_size - 1) // 2
-        self.video_shape = video_shape
+        self.dilation = cast_tuple(dilation, size = 3)
 
+        self.kernel_size = cast_tuple(kernel_size, size = 3)
+        assert all(map(lambda n: n % 2 == 1, self.kernel_size)), 'kernel size must be odd'
+
+        self.kernel_numel = self.kernel_size[0] * self.kernel_size[1] * self.kernel_size[2]
+
+        # calculate padding
+
+        self.padding_frame = calc_same_padding(self.kernel_size[0], self.dilation[0])
+        self.padding_height = calc_same_padding(self.kernel_size[1], self.dilation[1])
+        self.padding_width = calc_same_padding(self.kernel_size[2], self.dilation[2])
+
+        self.video_padding = (self.padding_width, self.padding_width, self.padding_height, self.padding_height, self.padding_frame, self.padding_frame)
+
+        # save video shape and calculate max number of tokens
+
+        self.video_shape = video_shape
         max_frames, fmap_size, _ = video_shape
         max_num_tokens = torch.empty(video_shape).numel()
         self.max_num_tokens = max_num_tokens
@@ -576,8 +594,8 @@ class Sparse3DNA(nn.Module):
 
         indices = torch.arange(max_num_tokens)
         shaped_indices = rearrange(indices, '(f h w) -> 1 1 f h w', f = max_frames, h = fmap_size, w = fmap_size)
-        padded_indices = F.pad(shaped_indices, (self.video_padding,) * 6, value = max_num_tokens) # padding has value of max tokens so to be masked out
-        unfolded_indices = unfoldNd(padded_indices, kernel_size = kernel_size, dilation = dilation)
+        padded_indices = F.pad(shaped_indices, self.video_padding, value = max_num_tokens) # padding has value of max tokens so to be masked out
+        unfolded_indices = unfoldNd(padded_indices, kernel_size = self.kernel_size, dilation = self.dilation)
         unfolded_indices = rearrange(unfolded_indices, '1 k n -> n k')
 
         # if causal, compare query and key indices and make sure past cannot see future
@@ -640,7 +658,7 @@ class Sparse3DNA(nn.Module):
         # reshape keys and values to video and add appropriate padding along all dimensions (frames, height, width)
 
         k, v = map(lambda t: rearrange(t, 'b (f h w) d -> b d f h w', f  = num_frames, h = fmap_size), (k, v))
-        k, v = map(lambda t: F.pad(t, (video_padding,) * 6), (k, v))
+        k, v = map(lambda t: F.pad(t, video_padding), (k, v))
 
         # put the attention processing code in a function
         # to allow for processing queries in chunks of frames
@@ -651,7 +669,7 @@ class Sparse3DNA(nn.Module):
             chunk_length = q.shape[1]
 
             k, v = map(lambda t: unfoldNd(t, kernel_size = kernel_size, dilation = dilation), (k, v))
-            k, v = map(lambda t: rearrange(t, 'b (d j) i -> b i j d', j = kernel_size ** 3), (k, v))
+            k, v = map(lambda t: rearrange(t, 'b (d j) i -> b i j d', j = self.kernel_numel), (k, v))
             k, v = map(lambda t: t[:, :chunk_length], (k, v))
 
             # append bos keys and values
@@ -696,8 +714,8 @@ class Sparse3DNA(nn.Module):
 
             # slice the keys and values to the appropriate frames, accounting for padding along frames dimension
 
-            kv_start_pos = ind
-            kv_end_pos = kv_start_pos + (ind + frames_per_chunk + video_padding * 2)
+            kv_start_pos = ind * frames_per_chunk
+            kv_end_pos = kv_start_pos + (ind + frames_per_chunk + self.padding_frame * 2)
             kv_frame_range = slice(kv_start_pos, kv_end_pos)
 
             k_slice, v_slice = map(lambda t: t[:, :, kv_frame_range], (k, v))
@@ -812,6 +830,90 @@ class Transformer(nn.Module):
 
         return self.norm(x)
 
+class ReversibleTransformer(nn.Module):
+    def __init__(
+        self,
+        *,
+        dim,
+        depth,
+        causal = False,
+        heads = 8,
+        dim_head = 64,
+        ff_mult = 4,
+        cross_attend = False,
+        attn_dropout = 0.,
+        ff_dropout = 0.,
+        ff_chunk_size = None,
+        sparse_3dna_attn = False,
+        sparse_3dna_kernel_size = 3,
+        sparse_3dna_video_shape = None,
+        sparse_3dna_query_num_frames_chunk = None,
+        sparse_3dna_dilations = (1,),
+        shift_video_tokens = False
+    ):
+        super().__init__()
+        assert not (sparse_3dna_attn and not exists(sparse_3dna_video_shape)), 'sparse_3dna_video_shape must be defined if turned on'
+        self.layers = MList([])
+
+        for ind in range(depth):
+            if sparse_3dna_attn:
+                dilation = sparse_3dna_dilations[ind % len(sparse_3dna_dilations)]
+                image_size = sparse_3dna_video_shape[-1]
+
+                self_attn = Sparse3DNA(
+                    dim = dim,
+                    heads = heads,
+                    dim_head = dim_head,
+                    causal = causal,
+                    kernel_size = sparse_3dna_kernel_size,
+                    dilation = dilation,
+                    video_shape = sparse_3dna_video_shape,
+                    query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
+                )
+            else:
+                image_size = None
+
+                self_attn = Attention(
+                    dim = dim,
+                    heads = heads,
+                    dim_head = dim_head,
+                    causal = causal,
+                    dropout = attn_dropout
+                )
+
+            wrapper_fn = partial(ShiftVideoTokens, image_size = image_size, shift_space = sparse_3dna_attn and shift_video_tokens)
+
+            self.layers.append(MList([
+                SandwichNorm(dim = dim, fn = wrapper_fn(self_attn)),
+                SandwichNorm(dim = dim, fn = wrapper_fn(FeedForward(dim = dim, mult = ff_mult, dropout = ff_dropout, chunk_size = ff_chunk_size)))
+            ]))
+
+            if not cross_attend:
+                continue
+
+            self.layers.append(MList([
+                SandwichNorm(dim = dim, fn = Attention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                SandwichNorm(dim = dim, fn = wrapper_fn(FeedForward(dim = dim, mult = ff_mult, dropout = ff_dropout, chunk_size = ff_chunk_size)))
+            ]))
+
+        attn_context_layer = ((True, False),) if cross_attend else tuple()
+        route_attn = ((True, False), *attn_context_layer) * depth
+        route_context = ((False, False), *attn_context_layer) * depth
+
+        context_route_map = {'context': route_context, 'context_mask': route_context} if cross_attend else {}
+        attn_route_map = {'mask': route_attn}
+
+        self.net = ReversibleSequence(self.layers, args_route = {**context_route_map, **attn_route_map})
+        self.norm = StableLayerNorm(dim)
+
+    def forward(
+        self,
+        x,
+        **kwargs
+    ):
+        x = self.net(x, **kwargs)
+        return self.norm(x)
+
 # embeddings
 
 class Embedding(nn.Module):
@@ -876,9 +978,11 @@ class NUWA(nn.Module):
         text_enc_depth = 6,
         text_enc_dim_head = 64,
         text_enc_heads = 8,
+        enc_reversible = False,
         dec_depth = 6,
         dec_dim_head = 64,
         dec_heads = 8,
+        dec_reversible = False,
         attn_dropout = 0.,
         ff_dropout = 0.,
         ff_chunk_size = None,
@@ -897,7 +1001,9 @@ class NUWA(nn.Module):
         self.text_embedding = Embedding(text_num_tokens, dim, frac_gradient = embed_gradient_frac)
         self.text_pos_embedding = Embedding(text_max_seq_len, dim)
 
-        self.text_transformer = Transformer(
+        enc_transformer_klass = Transformer if not enc_reversible else ReversibleTransformer
+
+        self.text_transformer = enc_transformer_klass(
             dim = dim,
             depth = text_enc_depth,
             heads = text_enc_heads,
@@ -921,7 +1027,9 @@ class NUWA(nn.Module):
 
         sparse_3dna_dilations = tuple(range(1, sparse_3dna_dilation + 1)) if not isinstance(sparse_3dna_dilation, (list, tuple)) else sparse_3dna_dilation
 
-        self.video_transformer = Transformer(
+        dec_transformer_klass = Transformer if not enc_reversible else ReversibleTransformer
+
+        self.video_transformer = dec_transformer_klass(
             dim = dim,
             depth = dec_depth,
             heads = dec_heads,
